/************************************************
 * Filename: ulti.c
 * SccsId:   @(#)ulti.c	1.4 94/12/08
 *
 ************************************************/

/*      ulti.c  V1.02	8.8.90 AJ mc HD		*/
/*
 *
 * Copyright (C) Parsytec GmbH, 1989.
 *
 * Author:	Matthias Clauss
 *		Design assistents by Armin Joachimsmeyer
 *
 * Date:	1.8.1989	Created
 *
 * Version:	1.02
 *
 * Updates:	Vers	date	author	description
 *		1.01	900626	mc	initialisation scheme changed
 *					additional support for mtm-sun
 *		1.02	900808	HD	more support for mtm-sun, no
 *					configuration if board in use
 *
 *	The ultimate transputer test-utility
 * Standalone utility for PARSYTEC VME Bus based
 * transputer modules.
 * First :
 *	If the '-v' option is set,
 *	display the hardware map as described in
 *	the local config file.
 * Note:
 *	Usually the next two steps are repeated for all
 *	existing transputers. Iff the '-l' option is present,
 *	it is done only for the specified transputer.
 * Second :
 *	Check if linkadaptors are really accessible
 *	on the VME Bus ie. if the the specified
 *	transputer board is really plugged in
 *	and the adresses are ok.
 * Third :
 *	Reset the tranputer connected to the linkadaptor
 *	and load a test program to check if the transputer
 *	is really connected and ready to work.
 *	on the VMTM eg this means that the C004 linkswitch
 *	has been correctly initialised for section B - D.
 */

#include <stdio.h>
#ifdef SunOS5         /* Solaris      */
#     include <unistd.h>
#else                 /* SunOS 4.1.x */
#     include <vfork.h>
#endif
#include <signal.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "../link_c/link.h"


static char *SccsId = "@(#)ulti.c	1.1 dated 5/19/92";	/* Sccs Id */


#define	DEFAULT_TEST_PRGM		"/usr/etc/transp/trans.cde"
#define	DEFAULT_VMTM_CONFIG_PNAME	"/usr/etc/transp/vmtmconf"
#define	DEFAULT_MTM_CONFIG_PNAME	"/usr/etc/transp/mtmconf"

#define	ENV_VMTM	"TRANSP_VMTM_CONF"
#define	ENV_MTM		"TRANSP_MTM_CONF"
#define	BUFF_SIZE	(4 * 1024)
#define	ERR_BUF_LEN	256	/* length of own error message buffer */
#define	BOOT_TIMEOUT	10	/* in units of 0.1 seconds */
#define	WRITE_TIMEOUT	10	/* in units of 0.1 seconds */
#define	READ_TIMEOUT	10	/* in units of 0.1 seconds */
#define	SIGUSR1_TIMEOUT	60	/* in units of 1 second    */
#define	EEXEC		1	/* child failed to exec configuring prgm */


/* external function declarations */

extern	int OpenLink  ();	/* (Name); */
extern	int CloseLink ();	/* (LinkId); */
extern	int ReadLink  ();	/* (LinkId, Buffer, Count, Timeout); */
extern	int WriteLink ();	/* (LinkId, Buffer, Count, Timeout); */
extern	int ResetLink ();	/* (LinkId); */
extern	int AnalyseLink ();	/* (LinkId); */
extern	int TestError ();	/* (LinkId); */
extern	int TestRead  ();	/* (LinkId); */
extern	int TestWrite ();	/* (LinkId); */
extern	int InfoLink  ();	/* (struct ext_link_info *); */


/* local variables */
static struct ext_link_info *x_link = NULL;
static int no_links = -1;   	/* number of links */
static char *vmtm_conf_pname;	/* program name to configure VMTMs */
static char *mtm_conf_pname;	/* program name to configure MTM-SUNs */
static int   sigusr1_received;	/* flag: SIGUSR1 caught */
static int   timeout_occured;	/* no signal caught (see SIGUSR1_TIMEOUT) */
static int   rflag;		/* flag: read existing topology */

/* functions declared inside this file */
void usage ();		/* display program usage */
void hello ();		/* welcome message */
int   show_info ();	/* display detailed info about transputer links */
int  test ();		/* dispatcher for testing transputer */
int  alive_and_test ();	/* check wether single transputer is alive */
int  code_load ();	/* load down a transputer binary executable */
void configure ();	/* dispatcher for configuring the topology */
int  config_board ();	/* configure (establish topology) on VMTM or MTM-SUN */
void sigusr_handler();	/* handler for SIGUSR1 signals */
void sigalrm_handler();	/* handler for SIGALRM signals */
int  name_to_id ();	/* convert a link-name into an unique link-id */

int
main (argc, argv, env)
int argc;
char *argv[], *env[];
{	int vflag, cflag, err;
	int i;
	char *ch;
	char *test_file, *conf_file, *link_name;
	register struct ext_link_info *x_l;

	/* external functions */
	void usage ();
	void hello ();
	int show_info ();
	int test ();
	void configure ();

	vflag = 0;
	cflag = 0;
	rflag = 0;
	link_name = (char *) 0;
	test_file = DEFAULT_TEST_PRGM;
	vmtm_conf_pname = DEFAULT_VMTM_CONFIG_PNAME;
	mtm_conf_pname  = DEFAULT_MTM_CONFIG_PNAME;

	/* process argv */
	for (i=1; i < argc;) {
		/* check command line arguments for
		 * '-l' , '-v' , '-t' or '-c' option
		 */
		ch = argv[i++];
		if ( *ch++ == '-' ) {
		    while (*ch) {
		        switch (*ch++) {
			
			    case 'c' :  cflag++;
			                break;
			    case 'l' :  if  ( i < argc ) {
					    link_name = argv[i++];
				        }
				        break;
			    case 'r' :  rflag++;
					cflag++;
			                break;
			    case 't' :  if  ( i < argc ) {
				            test_file = argv[i++];
				        }
				        break;
			    case 'v' :  vflag++;
			                break;
			    default:    usage ();
			                exit (1);
		        }	/* switch */
		    }	/* while */
		}
	} /* for */

	if ((no_links = InfoLink (&x_link)) < 0) {
		/* failed to get info about links */
		if (no_links == ESEM2GET) {
		    fprintf(stderr, "Can not open existing sem_key. Please \
check\ngroup and number of sems using \"ipcs -sb\":\n\n");
		    system("ipcs -sb");
		} else {
		    fprintf(stderr,
			"\nulti: Failed to get info about links, because: %s\n",
			lerror (no_links));
	        }
		return (1);
	}

	if ( vflag ) {
		hello ();
	}
	i = show_info(vflag);

	if ( cflag ) {
	    if (i != 0) {
		if (rflag)
		    fprintf(stderr, 
		     "can not read configuration, because still in use!\n\n");
		else
		    fprintf(stderr, 
		     "board(s) NOT reconfigured, because still in use!\n\n");
	    } else {
		configure (link_name, vflag);
	    }
	}

	if (rflag)
		return(0);
	return ( test (test_file, link_name, vflag) );
}



void usage ()
{
   printf ("\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
   "usage: ulti [-v] [-c] [-r] [-t test-program] [-l link-name]",
   "  configure (set topology) and test accessability of Parsytec transputers",
   "       -v verbose output",
   "       -c configure all boards (VMTM, MTM-SUN) with default topology",
   "       -r read existing configuration and display it",
   "       -t specifies other test-program (transputer binary)",
   "       -l configure/test only this link and not all");
}



void hello ()
{
	printf ("\n%s\n\n%s\n%s\n",
	   "Hi there! Welcome to the ultimate transputer board test-utility.",
	   "The config file tells me the following configuration",
	   "is available on your workstation.");
}



int show_info (showit)
int showit;
{	int i;
	int noofusedlinks = 0;
	register struct ext_link_info *x_l;

	if (showit)
		printf ("\n%s\n%s\n",
			" Id   Status       Link-Name         Bus-Address",
			"================================================");

	/* display info about each link */
	for (i=0, x_l=x_link; i < no_links; i++,x_l++) {
	    if (showit)
	    	if (strncmp(x_l->link_name, "scsi-t", 6) == 0) {
		    printf ("%3d   %-10s   %-16s   %s\n",
		        x_l->link_number,
		        (x_l->in_use==0) ? "available" : "IN USE",
		        x_l->link_name, (char*)  x_l->link_address);
	    	} else {
		    printf ("%3d   %-10s   %-16s   0x%08x\n",
		        x_l->link_number,
		        (x_l->in_use==0) ? "available" : "IN USE",
		        x_l->link_name, (unsigned long) x_l->link_address);
	    	}

		if (x_l->in_use)
			noofusedlinks++;
	}
	if (showit)
		printf ("\nnumber of used links: %d\n\n", noofusedlinks);
	return(noofusedlinks);
}

/*
 *	test
 *
 *	Test all or only single transputer. This is done by
 *	opening the according link, and sending down a test-program.
 *	This is an executable binary program for the transputer.
 *	It receives a single byte and sends it back incremented
 *	by one.
 *	Here only the handling wether a single or all links should
 *	be tested is done. For testing the function 'alive_and_test'
 *	is called.
 *
 *	Global variables
 *
 *	int link_no;	/* number of links known * /
 *	struct ext_link_info x_link[];
 */


test (test_file, link_name, verbose)
char *test_file;
char *link_name;
int   verbose;
{	int i;
	int tested, err, total;
	register struct ext_link_info *x_l;

	err   = 0;
	total = no_links;	/* number of links to be checked */
	if ( link_name ) {
		/* have to check only single transputer */
		err = alive_and_test (link_name, test_file, verbose);
	} else {
		/* test all links */
		tested = 0;
		for (i=0, x_l=x_link; i < no_links; i++, x_l++) {
		   /* an MTM-SUN config-link is not intended to be tested */
		   if ( (x_l->config_type == CONF_MTMSUN1) ||
			(x_l->config_type == CONF_MTMSUN2)) {
			total--;
			/* additionally skip testing */
			continue;
		   }

		   if ( x_l->in_use == 0 ) {
			/* link seems unused */
			if ( alive_and_test (x_l->link_name,
					     test_file, verbose) ) {
			   /* test failed */
			   err++;
			} else {
			   /* test passed ok */
			   tested++;
			}
		   } else {
			/* print diagnostic */
			printf ("Transputer link \"%s\" currently in use\n",
					x_l->link_name);
		   } /* else in_use */
		} /* for */
		if ( verbose ) {
		   printf ("\nHave tested %d of %d transputers successfully\n",
			tested, total);
		}
	}
	/* return error indicator */
	return (err);
}


/*
 *	alive_and_test
 *
 *	Process the link with the given name.
 *	First try to open the link. This involves by the link-interface
 *	an check for accessability on the VME bus.
 *	If the open succeeds, try to download a small check-program to
 *	the transputer. Then send a single byte and expect it back incremented
 *	by one (the transputer does this).
 *	Close the link or report any found errors.
 *	If all links are to processed, go on.
 *	Return 0 iff no problem deteced, otherwise 1.
 */


int alive_and_test (link_name, test_file, verbose)
char *link_name;
char *test_file;
int   verbose;
{	int link_id;
	int lerr;			/* local error code */
	unsigned char sdata, rdata;	/* send/receive-data for transputer */
	char err_buf[ERR_BUF_LEN];	/* error message is collected here */
	int add_errno_text;		/* if != 0 append 'perror' msg */
	extern int errno, sys_nerr;
	extern char *sys_errlist[];	/* for system error messages */

	add_errno_text = FALSE;

	if (verbose) {
		/* announce test */
		printf ("Testing link \"%s\" ... ", link_name);
		fflush (stdout);
	}

	/* try to open the link */
	link_id = OpenLink (link_name);
	if ( link_id < 0 ) {
		/* print diagnostic */
		if (link_id == ESEM2GET) {
		    sprintf(err_buf, "Can not open existing sem_key. Please \
check\ngroup and number of sems using \"ipcs -sb\":\n\n");
			system("ipcs -sb");
		} else {
			sprintf(err_buf, "Failed to open link, because: %s\n",
				lerror (link_id));
		}
		goto error;
	}

	/* download test program to transputer */
	lerr = code_load (link_id, test_file, err_buf);
	if ( lerr < 0 ) {
		/* add system error message to err_buf */
		add_errno_text = TRUE;
		goto error;
	}
	if ( lerr > 0 ) {
		/* no error post-processing needed */
		goto error;
	}

	/* send single byte to transputer */
	lerr = WriteLink (link_id, &sdata, 1, WRITE_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during send */
		sprintf (err_buf,
			"Failed to send single byte 0x%x, because: %s\n",
			sdata, (lerr == 0) ? "Timeout" : lerror (lerr));
		goto error;
	}

	/* receive byte incremented by one from transputer */
	lerr = ReadLink (link_id, &rdata, 1, READ_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during receive */
		sprintf (err_buf, "Failed to receive answer, because: %s\n",
			(lerr == 0) ? "Timeout" : lerror (lerr));
		goto error;
	}

	/* check that send++ == received */
	if ( sdata+1 != rdata ) {
		/* answer is wrong */
		sprintf (err_buf, "Have received wrong answer 0x%x/0x%x\n",
			(unsigned) rdata, (unsigned) sdata);
		goto error;
	}

	/* everything is perfectly fine. Close link */
	CloseLink (link_id);
	if (verbose) {
		/* all tests succeed */
		printf ("OK\n");
		fflush (stdout);
	}
	return (0);

error:
	/* some kind of error has happend. */
	CloseLink (link_id);
	if ( !verbose ) {
		printf ("Testing transputer link \"%s\" has FAILED\n",
			link_name);
		fflush (stdout);
		return (1);
	}
	/* verbose option set */
	printf ("FAILED!\n     %s", err_buf);
	if ( add_errno_text ) {
		/* additionally print system error message */
		if ( (0 < errno) && (errno <= sys_nerr) ) {
			printf ("%s (errno %d)\n", sys_errlist[errno], errno);
		} else {
			/* errno out of valid range */
			printf ("errno (%d) out of valid range\n", errno);
		}
	}
	fflush (stdout);
	return (1);
}


/*
 *	code_load
 *
 *	Assumes the link is opened. Resets transputer and send down
 *	the 'boot_file'. Error messages will be written in 'err_buf'.
 *	Returns 0 iff everything is ok, +1 if error but no more error
 *	handling needed by caller or -1 if system error message
 *	should be added.
 */



int code_load (link_id, boot_file, err_buf)
int   link_id;
char *boot_file;
char *err_buf;
{	int lerr;	/* local error flag */
	FILE *fp;	/* boot file */
	int cnt;
	int total_send;			/* bytes send so far */
	char fbuffer[BUFF_SIZE];	/* test-prgm file buffer */

	/* reset transputer */
	lerr = ResetLink (link_id);
	if ( lerr < 0 ) {
		/* failed to reset link */
		sprintf (err_buf, "Failed to reset link, because: %s\n",
			lerror (lerr));
		return (1);
	}

	/* download transputer code-file */
	fp = fopen (boot_file, "r");
	if ( fp == NULL ) {
		/* failed to open boot file */
		sprintf (err_buf, "Failed to open file \"%s\", because: ",
			boot_file);
		return (-1);
	}
	/* boot file opened successfully */
	total_send = 0;
	while ((cnt = fread (fbuffer, 1, sizeof (fbuffer), fp)) > 0) {
		/* try to send fbuffer to transputer */
		lerr = WriteLink (link_id, fbuffer, cnt, BOOT_TIMEOUT);
		if ( lerr != cnt ) {
			/* failed to send complete data */
			if ( lerr == 0 ) {
			   /* probably not configured */
			   sprintf (err_buf,
				"%s\n",
		"Link not connected to a transputer (board not configured)");
			   return (1);
			}
			if ( lerr > 0 ) {
			   /* timeout during send */
			   sprintf (err_buf,
				"Timeout after %d from %d bytes send\n",
				lerr, cnt);
			   return (1);
			} else {
			   /* failed to send */
			   sprintf (err_buf, 
				"Failed to send byte 0x%X, because: %s\n",
				total_send, lerror (lerr));
			   return (1);
			}
		}
		total_send += cnt;
	} /* while */
	/* check for fread error */
	if ( feof (fp) ) {
		/* end of file */
		fclose (fp);
	} else {
		/* error during read */
		sprintf (err_buf, "File read error on \"%s\", because:",
			boot_file);
		return (-1);
	}

	return (0);
}


/*
 *	configure
 *
 *	configure all vmtm and mtm-sun transputer boards.
 *
 *	The procedure is as follows:
 *	If 'link_name' != NULL, work on this link only. Otherwise
 *	deal with all link, which needs configuration.
 *	Collect this information from the global 'x_link' array.
 *	The struct component 'may_config' contains the board type
 *	for configuration. The full link-name is passed down to
 *	an configuring program. The programs names are passed as
 *	global variables 'vmtm_conf_pname' and 'mtm_conf_pname'.
 *	Each prgm is 'vfork()'-ed and 'execvp(3)'-ed.
 *	The vmtm configuring prgm is expected to return the return
 *	value 0 and 'wait()'-ed for.
 *	The mtm-sun configuring prgm is expected to send SIGUSR1
 *	anyway, while printing error messages to stderr. This is due
 *	to the future mtm-sun ncm, which will exist even after the
 *	initialisation.
 *	If 'verbose' print a lot of informative messages.
 *	Anyway print error messages.
 *
 *	Global variables
 *	struct ext_link_info[]	x_link;	   RO
 *	int			no_links;  RO
 *	char   *vmtm_conf_pname;
 *	char   *mtm_conf_pname;
 *	
 */


void configure (link_name, verbose)
char *link_name;
int   verbose;
{	int i;
	int l_id;	/* tmp link_id */
	register struct ext_link_info *x_l;

	/* functions called from this function */
	int name_to_id ();
	char *getenv ();

	/* try to overwrite default settings for configuring
	 * program names for VMTM and MTM-SUN.
	 */
	vmtm_conf_pname = getenv (ENV_VMTM);
	if ( vmtm_conf_pname == NULL ) {
	    /* failed to get filename of VMTM configuring
	     * program from the environment.
	     * Use default config-file name
	     */
	    vmtm_conf_pname = DEFAULT_VMTM_CONFIG_PNAME;
	}

	mtm_conf_pname = getenv (ENV_MTM);
	if ( mtm_conf_pname == NULL ) {
	    /* failed to get filename of MTM-SUN configuring
	     * program from the environment.
	     * Use default config-file name
	     */
	    mtm_conf_pname = DEFAULT_MTM_CONFIG_PNAME;
	}

	if ( link_name ) {
		/* configure the single given link */
		/* what a pitty. First I must convert the user supplied
		 * 'link_name' into a standard link-name and check,
		 * wether it needs to be configured and if so, in
		 * which way. Easy but expensive is OpenLink() and
		 * then directly CloseLink() to get the unique
		 * identifying LinkId. 'name_to_id()' does this.
		 */
		if ( (l_id = name_to_id (link_name) ) < 0) {
			/* failed to convert the user supplied link_name
			 * into a link_id. Fatal, abort.
			 */
			printf ("Failed to recognize link \"%s\"\n",
				link_name);
		} else {
		   x_l = &x_link[l_id];
		   /* check wether link seems to be free */
		   if ( x_l->in_use == 0 ) {
			/* link seems unused */
			config_board (x_l, verbose);
		   } else {
			/* print diagnostic */
			printf ("Transputer link \"%s\" currently in use\n",
					x_l->link_name);
		   } /* else in_use */
		   /* configure the single link 'link_name'. */
		   config_board ( x_l, verbose);
		} /* else name_to_id */
	} else {
		/* configure all links which need that */
		for (i=0, x_l=x_link; i < no_links; i++, x_l++) {
		   if ( x_l->config_type <= 0 ) {
			continue;
		   }
		   if ( x_l->in_use == 0 ) {
			/* link seems unused */
			config_board (x_l, verbose);
		   } else {
			/* print diagnostic */
			printf ("Transputer link \"%s\" currently in use\n",
					x_l->link_name);
		   } /* else in_use */
		}
	} /* else link_name */
}


/*
 *	config_board
 *
 *	send configuration program to single transputer
 *
 *	Global variables
 *
 *	struct ext_link_info	x_link[];	/* check wether the given
 *						 * link may be configured.
 *						 * /
 */


int config_board (link_desc, verbose)
struct ext_link_info *link_desc;
int   verbose;
{	int lerr;			/* local error code */
	int config_type;		/* board type for configuration */
	char err_buf[ERR_BUF_LEN];	/* error message is collected here */
	int add_errno_text;		/* if != 0 append 'perror' msg */
	int pid, wpid;			/* pid of child, wpid 'wait(2) */
#ifdef SunOS5           /* Solaris       */
        int       *wstat;
#else                   /* SunOS  4.1.x  */
        union wait wstat;               /* 'wait(2)' status info */
#endif
	struct itimerval usr1_time;	/* for SIGUSR1 timeout */

	extern int errno, sys_nerr;
	extern char *sys_errlist[];	/* for system error messages */

	/* functions referenced from here */
	void sigusr_handler ();		/* handler for SIGUSR1 and SIGUSR2 */
	void sigalrm_handler ();	/* SIGALRM signal handler */

	add_errno_text = FALSE;
	/* setup timeout value for SIGUSR1 signal */
	usr1_time.it_interval.tv_sec  = 0L; /* no automatic reload */
	usr1_time.it_interval.tv_usec = 0L;
	usr1_time.it_value.tv_sec  = SIGUSR1_TIMEOUT;
	usr1_time.it_value.tv_usec = 0L;

	if (verbose) {
		/* announce test */
		printf ("\n\nConfiguring link \"%s\" ...\n", 
			link_desc->link_name);
		fflush (stdout);
	}

	config_type = link_desc->config_type;

	/* checked wether this link may be configured */
	if ( config_type <= CONF_NULL ) {
		/* print diagnostic */
		sprintf (err_buf, "This link must not be configured\n");
		goto error;
	}

	/* checked for unknown configuration link */
	if ( config_type > CONF_LAST ) {
		/* print diagnostic */
		sprintf (err_buf, "Link has unknown \'config_type\' (%d)\n",
			config_type);
		goto error;
	}

	/* check for vmtm board */
	if ( config_type == CONF_VMTM ) {
		/* configure a VMTM board
		 *
		 * Since V1.01 the strategy has changed completly.
		 * Now a process is 'vfork()'-ed and 'execlp(3)'
		 * the vmtm-configuring program. Its name is given
		 * by the global variable 'vmtm_conf_pname'. Its
		 * value may be changed from the default one by an
		 * environment variable.
		 * The program is passed as a single argument the full
		 * link-name. An exit-value of 0 indicates successfull
		 * configuration, so that VMTM configuring program is
		 * 'wait4()'-ed for.
		 */
		if ( (pid = vfork()) != 0 ) {
		   /* in parent process here.
		    * wait for child
		    */
#ifdef SunOS5           /* Solaris      */
                   if ( (wpid = wait ( wstat )) == -1 ) {
#else                   /* SunOS 4.1.x */
		   if ( (wpid = wait ( &wstat )) == -1 ) {
#endif
			/* wait for child has failed */
			sprintf (err_buf, 
				 "wait for child (%d) failed, errno=%d\n",
				pid, errno);
			goto error;
		   } else {
			/* wait() returns */
			if ( wpid != pid ) {
			   /* wait reports wrong child */
			   sprintf (err_buf,
			      "wait returns pid=%d, expected pid=%d\n",
			      wpid, pid);
			   goto error;
			}
#ifdef SunOS5   /* Solaris      */
                        if ( WIFEXITED(*wstat) ) {
                           /* wait report of child due to exit */
                           if ( WEXITSTATUS(*wstat) != 0 ) {
                                /* VMTM config prgm returns exit-value
                                 * != 0.
                                 */
                                sprintf (err_buf,
                                   "program \"%s\" returns %d\n",
                                     vmtm_conf_pname, WEXITSTATUS(*wstat) );

#else           /* SunOS 4.1.x */
			if ( WIFEXITED(wstat) ) {
			   /* wait report of child due to exit */
			   if ( wstat.w_retcode != 0 ) {
				/* VMTM config prgm returns exit-value
				 * != 0.
				 */
				sprintf (err_buf,
				   "program \"%s\" returns %d\n",
				   vmtm_conf_pname, wstat.w_retcode);
#endif          /* SunOS5       */

				goto error;
			   } else {
			      /* everything fine, fall through to XXX */
			      ;
			   } /* else w_retcode != 0 */
			} else {
			   /* wait reports child not exited */
			   /* perhaps give more diagnostic infos */
			   sprintf (err_buf, 
				    "child has not exited, please report\n");
			   goto error;
			} /* else WIFEXITED */
		   } /* else wait() == -1 */
		   /* end of parent code. It continues after the
		    * else part for the 'vfork()'-ed child.
		    */
		} else {
		   /* in 'vfork()'-ed child here.
		    * Try to 'execlp(3)' the VMTM configuring program.
		    */
		   if ( execlp (vmtm_conf_pname, vmtm_conf_pname,
				link_desc->link_name, 
				verbose? "-v": (rflag? "-r" : NULL),
				rflag? "-r": NULL,
				(char *)NULL) == -1 ) {
			/* executing the configuring VMTM program failed */
			fprintf (stderr, "Failed to execute \"%s\" errno=%d\n",
				vmtm_conf_pname, errno);
			_exit (EEXEC);
		    }
		    /* end of child: have successfully executed VMTM
		     * configuring program.
		     */
		} /* else vfork() != 0 */
	/* else config_type == CONF_VMTM */
	} else if ( (config_type == CONF_MTMSUN1) ||
		    (config_type == CONF_MTMSUN2) ) {
		/* configure a MTM-SUN board.
		 * 
		 * First have a glance on the strategy:
		 * A process is 'vfork()'-ed and 'execlp(3)'
		 * the mtm-configuring program. Its name is given
		 * by the global variable 'mtm_conf_pname'. Its
		 * value may be changed from the default one by an
		 * environment variable.
		 * The program is passed as a single argument the full
		 * link-name. This program may load down a ncm and act as
		 * interface for requests from the sun. Therefor it is
		 * not expected to terminate. On the other hand there
		 * should be a possibility, to signal successfull ini-
		 * tialisation to this program. So, here is what I do:
		 * After `exec`-ing the MTM-SUN configuring program
		 * the SIGUSR1 or SIGUSR2 signal is expected to. SIGUSR1
		 * means SUCCESS, SIGUSR2 means FAILURE. Then the rest of
		 * the links are searched for boards, that need to be
		 * configured.
		 */
		/* the parent must install a signal-handler for
		 * SIGUSR1 and SIGUSR2, because these signals cause process
		 * termination on default.
		 * Unfortunatly the parent is blocked after a 'vfork()' call,
		 * so I must install the handler here.
		 */
		if ( signal (SIGUSR1, sigusr_handler) == (void *)-1 ) {
		   /* cannot install signal handler for SIGUSR1 */
		   sprintf(err_buf, 
			   "cannot install SIGUSR1 handler, errno=%d\n",
			errno);
		   goto error;
		}
		if ( signal (SIGUSR2, sigusr_handler) == (void *)-1 ) {
		   /* cannot install signal handler for SIGUSR2 */
		   sprintf(err_buf, 
			   "cannot install SIGUSR2 handler, errno=%d\n",
			errno);
		   goto error;
		}
		/* clear flag SIGUSR1 received */
		sigusr1_received = FALSE;
		timeout_occured  = FALSE;
		if ( (pid = vfork()) != 0 ) {
		   /* in parent process here.
		    * awaiting signal SIGUSR1 from child
		    * rest to sleep for a while, until either
		    * timeout occurs or SIGUSR1 from child arrives.
		    */
		   if ( setitimer (ITIMER_REAL, &usr1_time, 
				    (struct itimerval *)NULL) 
			== -1 ) {
			    /* cannot set itimer for SIGUSR1 timeout */
			    sprintf(err_buf, 
				    "cannot set USR1 timeout, errno=%d\n", 
				    errno);
			    goto error;
		   }
		   if ( signal (SIGALRM, sigalrm_handler) == (void *)-1 ) {
			 /* cannot install signal handler for SIGALRM */
			 sprintf(err_buf, 
				 "cannot install SIGALRM handler, errno=%d\n",
				 errno);
			 goto error;
		   }

		   /* release cpu without blocking any signals */
		   sigpause (0);
		   /* check wether MTM-SUN configuring program
		    * has send SIGUSR1 or the timeout condition
		    * causes 'sigpause()' to return.
		    */
		   if ( sigusr1_received ) {
			/* MTM-SUN configuring program has been
			 * started successfully.
			 */
			;
		   } else if (timeout_occured) {
			/* no SIGUSR1 signal received from
			 * MTM-SUN configuring program
			 * within SIGUSR1_TIMEOUT seconds.
			 * report warning
			 */
			sprintf (err_buf,
			    "timeout: missing SIGUSR1 signal from prgm \"%s\" (pid=%d)\n",
			    mtm_conf_pname, pid);
			goto error;
		   } else {
			/* SIGUSR2 received from MTM-SUN configuring program.
			 * report warning
			 */
			sprintf (err_buf, "Error: configuration failed\n");
			goto error;
		   }
		   /* end of parent code. It continues after the
		    * else part for the 'vfork()'-ed child.
		    */
		} else {
		   /* in 'vfork()'-ed child here.
		    * Try to 'execlp(3)' the VMTM configuring program.
		    */
		   if ( execlp (mtm_conf_pname, mtm_conf_pname,
				link_desc->link_name, 
				verbose? "-v": (rflag? "-r": NULL),
				rflag? "-r": NULL,
				(char *)NULL) == -1 ) {
			/* executing the configuring MTM-SUN program failed */
			fprintf (stderr, "Failed to execute \"%s\" errno=%d\n",
				mtm_conf_pname, errno);
			_exit (EEXEC);
		    }
		    /* end of child: have successfully executed MTM-SUN
		     * configuring program.
		     */
		} /* else vfork() != 0 */
	/* else config_type == (CONF_MTMSUN1) || (CONF_MTMSUN2) */
	}
	if (verbose) {
		/* all tests succeed */
		printf ("Configuring link \"%s\" has SUCCEEDED\n\n", 
			link_desc->link_name);
		fflush (stdout);
	}
	return (0);

error:
	/* some kind of error has happend. */
	printf ("Configuring transputer link \"%s\" has FAILED\n\n",
		link_desc->link_name);
	if (verbose) {
		printf ("%s", err_buf);
		if ( add_errno_text ) {
			/* additionally print system error message */
			if ( (0 < errno) && (errno <= sys_nerr) ) {
				printf ("%s (errno %d)\n", sys_errlist[errno],
					errno);
			} else {
				/* errno out of valid range */
				printf ("errno (%d) out of valid range\n", 
					errno);
			}
		}
	}
	fflush (stdout);
	return (1);
}


/*
 *	sigusr_handler
 *
 *	Signal handler for signal SIGUSR1 and SIGUSR2.
 *	When catching a SIGUSR1 or SIGUSR2 signal this function will be called
 *	by SunOs. It simply sets the global vcariable 'sigusr1_received'
 *	to TRUE, so that the functyion 'config_board()' may detect in a
 *	secure way, that the configuration of the MTM-SUN has succeeded.
 *	Receiving a SIGUSR2 means that there were some errors in the
 *	configuration.
 *
 *	Global variables
 *
 *	int sigusr1_received;	WO
 */


void sigusr_handler (sig)
int sig;
{
	/* check wether calling event was SIGUSR1 */
	sigusr1_received = ( sig == SIGUSR1 );
}


/*
 *	sigalrm_handler
 *
 *	SIgnal handler for signal SIGALRM.
 *	When catching a SIGALRM signal this function will be called
 *	by SunOs. It sets the global variable 'timeout_occured' to
 *	TRUE.
 */

void sigalrm_handler (sig)
int sig;
{
	/* check wether calling event was SIGALRM */
	if ( sig != SIGALRM ) {
		/* handler was called not for signal SIGUSR1.
		 * I do not have any idea under which circumstances
		 * this may happen, it is a fire door.
		 */
		fprintf (stderr, "\nulti: Caught unexpected signal %d\n", sig);
	} else {
		timeout_occured = TRUE;
	}
}


/*
 *	name_to_id
 *
 *	Converts the passed 'link_name' into a LinkId.
 *	The conversion is expensive, by OpenLink()' directly followed
 *	by 'CloseLink()'.
 *	Returns the according LinkId (>=0) or the ErrorCode from
 *	'OpenLink()'.
 */


int name_to_id (link_name)
char *link_name;
{	int link_id = -1;	/* LinkId */

	if ( (link_id = OpenLink (link_name)) < 0 ) {
		/* link_name is not a valid name.
		 * Return the error-code.
		 */
		return (link_id);
	} else {
		/* have opened 'link_name' successfully.
		 * Close this link and return LinkId.
		 */
		CloseLink (link_id);
		return (link_id);
	}
}
