/*
 * lconf.c	Configuration for general interface to transputer hardware
 *
 * SCCSid:	see below
 *
 * Copyright (C) Parsytec GmbH, 1989.
 * Copyright (C) CWA mbH, 1991 (for all parts concerning SBUS/BBK-S4).
 *
 * Author:	Matthias Clauss
 *		Design assistents by Armin Joachimsmeyer
 *
 *		Michael Raus (all parts concerning SBUS/BBK-S4)
 *
 * Date:	1.8.1989	Created
 *
 * Version:	2.2
 *
 * Updates:	Vers	date	author	description
 *		1.01	3.8.89	mc	config file in sub-direct. 'transp'
 *		1.02	18.9.89	mc	Hint for Sun 4/110 machines
 *		1.03	290791	mr	support of SBus devices
 *		1.04	180592	mc	more explanations given
 *              1.1     210194  nvs     support for Solaris 2.x device scheme
 *              2.2	071294	bla	inserted switch for 24bit
 */

#include <sys/types.h>

/*
 * the following variables are defined here and exported to
 * 'link.o', the general transputer interface, in order to
 * offer the administrator the possibility to choose other
 * values for some interesting constants.
 * 
 * The transputer link-interface needs to know where the
 * configuration file is located. This files describes the
 * type (VMTM, BBK-S4) and address (VME-address or SBus-slot-no).
 * The default value is defined in 'config_file' below.
 * Iff there exists an environment variable named 'env_config',
 * its value is used instead. This offers a maximum of flexibility.
 */

static char *SccsId = "@(#)lconf.c	1.1	94/12/07";	/* Sccs Id */

char *config_file    = "/usr/etc/transp/trans_conf";
						/* default config-file name */
char *env_config     = "TRANSP_CONFIG";		/* enironment variable name
						 * which contains the name
						 * of transputer config-file
						 */
char *vme_group_name = "vmebus";		/* vmebus group name.
						 * This group is given the
						 * permission to rw to the
						 * VME/SBus-device.
						 * For historical reasons
						 * not changed for SBus.
						 */
/*
 *	VMEBus or SBus device
 *	for Sun 4/110 use "/dev/vme24d16",
 */
		/* The Sun4-110 is the only maschine
		 * with a 24Bit address-bus as VME-Bus.
		 * So install the transputer interfaces
		 * as 24Bit addressed devices using '/dev/vme24d16'
		 * instead of the standard '/dev/vme32d16'
		 */

#ifdef VME24BIT
char *vme_dev_name   = "/dev/vme24d16";		/* Sun4-110 */
#else
char *vme_dev_name   = "/dev/vme32d16";
#endif
				/* Hint for network installation:
				 *
				 * You may change this to '/dev/vmebus'.
				 * On every Sun create such an device with
				 *   'mknod /dev/vmebus c 3 [5-10]'.
				 * Use 'ls -lg /dev/vme*' to find the correct
				 * major (3) and minor ([5-10]) device numbers.
				 */



char *sbus_dev_name  = "/dev/sbus";
					/* The SBus slot number is added
					 * by the driver (link.o) himself!
					 * (Imagine two BBK-S4 installed!)
					 */
					
key_t sem_link_key   = ((key_t)0x42424242);
					/* semaphore transputer key */
int   default_tenth  = 88000;	/* default 0.1s loop count
				 * Will be used only iff automatic timeout
				 * measurement failes. (Never noticed this)
				 */
unsigned reset_raise = 20000;	/* ResetLink() reset raise time in usec */
unsigned reset_fall  = 20000;	/* ResetLink() reset fall  time in usec */
